/**
 *
 * Beryl gconf settings backend
 *
 * gconf.c
 *
 * Copyright (c) 2006 Robert Carr <racarr@beryl-project.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 **/



#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <libintl.h>
#include <string.h>

#include <beryl-settings-backend.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include <gconf/gconf-value.h>

#define GROUPKEY    gchar * groupname=setting->parent->name?setting->parent->name:"_"; \
            gchar * keyname=g_strconcat(setting->is_screen?"s_":"a_",setting->name,NULL)

#define PATH             g_strconcat("/apps/beryl/",groupname,"/",keyname,NULL)


GConfClient *client = NULL;

Bool backendTriggered = TRUE;

/* Bindings */


struct _Modifier
{
	gchar *name;
	gint modifier;
}
const static modifiers[] = {
	{"<Shift>", ShiftMask},
	{"<Control>", ControlMask},
	{"<Mod1>", Mod1Mask},
	{"<Mod2>", Mod2Mask},
	{"<Mod3>", Mod3Mask},
	{"<Mod4>", Mod4Mask},
	{"<Mod5>", Mod5Mask},
	{"<Alt>", CompAltMask},
	{"<Meta>", CompMetaMask},
	{"<Super>", CompSuperMask},
	{"<Hyper>", CompHyperMask},
	{"<ModeSwitch>", CompModeSwitchMask},
};

#define N_MODIFIERS (sizeof (modifiers) / sizeof (struct _Modifier))


static gchar *mods_to_string(unsigned int mods)
{
	gchar *retstr = g_strdup("");
	gchar *tmpstr = retstr;
	gint i;

	for (i = 0; i < N_MODIFIERS; i++)
	{
		if (mods & modifiers[i].modifier)
		{
			retstr = g_strconcat(retstr, modifiers[i].name, NULL);
			g_free(tmpstr);
			tmpstr = retstr;
		}
	}
	return retstr;
}

static gboolean get_binding_is_enabled(gchar * src)
{
    if (strcasecmp(src,"Disabled")==0 || !strlen(src))
        return FALSE;
    if (src[0]=='#')
        return FALSE;
    return TRUE;
}


static unsigned int get_mods_and_endptr(gchar * src, gchar ** ret)
{
    unsigned int mods=0;
    gchar * spos=src;
    while((spos=strchr(spos,'<')) && *src)
    {
        int i;
        for (i=0;i<N_MODIFIERS;i++)
            if (strncasecmp(modifiers[i].name,
                        spos,strlen(modifiers[i].name))==0)
            {
                mods|=modifiers[i].modifier;
                spos+=strlen(modifiers[i].name);
                src=spos;
                break;
            }
        if (i==N_MODIFIERS)
            break;
    }
    *ret=src;
    return mods;
}

static void set_key_binding_from_string(BerylSettingValue * value, gchar * src)
{
    gboolean enabled=get_binding_is_enabled(src);
    int keysym=0;
    int keymods=0;
    gchar * spos;
    keymods=get_mods_and_endptr(src,&spos);
    if (spos && *spos)
    {
        keysym=XStringToKeysym(spos);
    };
    beryl_setting_value_set_keysym(value,&keysym);
    beryl_setting_value_set_keymods(value,&keymods);
    beryl_setting_value_set_key_enabled(value,&enabled);
}

static void set_button_binding_from_string(BerylSettingValue * value, gchar * src)
{
    gboolean enabled=get_binding_is_enabled(src);
    int button=0;
    int buttonmods=0;
    gchar * spos;
    buttonmods=get_mods_and_endptr(src,&spos);
    if (spos && *spos)
    {
        spos=strcasestr(spos,"Button");
        if (spos && *spos)
        {
            spos+=strlen("Button");
            button=atoi(spos);
        }
    }
    beryl_setting_value_set_button(value,&button);
    beryl_setting_value_set_buttonmods(value,&buttonmods);
    beryl_setting_value_set_button_enabled(value,&enabled);
}




/* --- */

/* Colors */
static void set_array_from_color(BerylSettingValue * value, gint * array)
{
	memcpy(array, value->value.as_color.array.array, sizeof(gint) * 4);
}

static void copy_color(BerylSettingValue * value, gint ** ptr)
{
	set_array_from_color(value, *ptr);
	*ptr += 4;
}

static void set_color_value_from_array(BerylSettingValue * value,
									   gint * array)
{
	BerylSettingColorValue color_val;

	memcpy(color_val.array.array, array, sizeof(gint) * 4);
	beryl_setting_value_set_color(value, &color_val);
}

/* --- */

/* Reading */

static void readBool(BerylSetting * setting)
{
	GROUPKEY;
	GError *e = NULL;
	gboolean val = gconf_client_get_bool(client, PATH, &e);

	if (!e)
	{
		beryl_setting_value_set_bool(&setting->value, &val);
	}
	g_free(keyname);
}

static void readInt(BerylSetting * setting)
{
	GROUPKEY;
	GError *e = NULL;
	int val = gconf_client_get_int(client, PATH, &e);

	if (!e)
	{
		beryl_setting_value_set_int(&setting->value, &val);
	}
	g_free(keyname);
}

static void readFloat(BerylSetting * setting)
{
	GROUPKEY;
	GError *e = NULL;
	gdouble val = gconf_client_get_float(client, PATH, &e);

	if (!e)
	{
		beryl_setting_value_set_float(&setting->value, &val);
	}
	g_free(keyname);
}

static void readString(BerylSetting * setting)
{

	GROUPKEY;
	GError *e = NULL;
	gchar *val = gconf_client_get_string(client, PATH, &e);

	if (!e && val)
	{
		setting->value.value.as_string = g_strdup(val);
		setting->is_default = FALSE;

	}
	g_free(keyname);

}

static void readColor(BerylSetting * setting)
{
	GROUPKEY;
	gint *int_ret = malloc(sizeof(gint) * 4);
	gint *ptr = int_ret;

	*ptr = gconf_client_get_int(client, g_strconcat(PATH, "/red", NULL),
								NULL);
	ptr++;
	*ptr = gconf_client_get_int(client, g_strconcat(PATH, "/green", NULL),
								NULL);
	ptr++;
	*ptr = gconf_client_get_int(client, g_strconcat(PATH, "/blue", NULL),
								NULL);
	ptr++;
	*ptr = gconf_client_get_int(client, g_strconcat(PATH, "/alpha", NULL),
								NULL);
	set_color_value_from_array(&setting->value, int_ret);
	g_free(int_ret);
	g_free(keyname);

}


static void readBinding(BerylSetting * setting)
{
	GROUPKEY;
	setting->is_default=TRUE;
	gchar *nkeyname;
	gchar *str;
	int intv;
	gboolean boolv;
	nkeyname=g_strconcat(keyname,"__keyboard",NULL);
	GError *e;
	e=NULL;
	str=gconf_client_get_string(client,g_strconcat("/apps/beryl",groupname,"/",nkeyname,NULL),e);
	
	if (!e && str)
	{
		setting->is_default=FALSE;
		set_key_binding_from_string(&setting->value,str);
		g_free(str);
	}
	if (e)
	{
		g_error_free(e);
		e=NULL;
	}
	
	g_free(nkeyname);
	nkeyname=g_strconcat(keyname,"__mouse",NULL);
	str=gconf_client_get_string(client,g_strconcat("/apps/beryl",groupname,"/",nkeyname,NULL),e);
	
	if (!e && str)
	{
		setting->is_default=FALSE;
		set_button_binding_from_string(&setting->value,str);
		g_free(str);
	}
	if (e) 
	{
		g_error_free(e);
		e=NULL;
	}
	g_free(nkeyname);

	nkeyname=g_strconcat(keyname,"__edge",NULL);
	intv=gconf_client_get_int(client,g_strconcat("/apps/beryl",groupname,"/",nkeyname,NULL),e);
	if (!e)
	{
		beryl_setting_value_set_edgemask(&setting->value,&intv);
	}
	g_free(nkeyname);

	if (e)
	{
		g_error_free(e);
		e=NULL;
	}

	nkeyname=g_strconcat(keyname,"__bell",NULL);
	boolv=gconf_client_get_bool(client,g_strconcat("/apps/beryl",groupname,"/",nkeyname,NULL),e);
	if (!e)
	{
		setting->is_default=FALSE;
		beryl_setting_value_set_bell(&setting->value,&boolv);
	}
	if (e)
	{
		g_error_free(e);
		e=NULL;
	} 
	g_free(nkeyname);
}

static void readList(BerylSetting * setting)
{
	GROUPKEY;
	switch (setting->info.for_list.list_of_type)
	{
		case BERYL_SETTING_TYPE_STRING:
		{
			GSList *settingList = gconf_client_get_list(client,PATH,GCONF_VALUE_STRING,NULL);
			gchar * val;
			while (val = g_slist_next(settingList))
			{
		            BerylSettingValue * value=beryl_setting_list_append(setting);
                            value->value.as_string=g_strdup(val);
                            value->parent->is_default=FALSE;
			}
		}
	}
}

static void readSwitch(BerylSetting * setting)
{

	switch (setting->type)
	{
	case BERYL_SETTING_TYPE_BOOL:
		readBool(setting);
		break;
	case BERYL_SETTING_TYPE_INT:
		readInt(setting);
		break;
	case BERYL_SETTING_TYPE_FLOAT:
		readFloat(setting);
		break;
	case BERYL_SETTING_TYPE_STRING:
		readString(setting);
		break;
	case BERYL_SETTING_TYPE_COLOR:
		readColor(setting);
		break;
	case BERYL_SETTING_TYPE_BINDING:
		readBinding(setting);
		break;
	case BERYL_SETTING_TYPE_LIST:
		readList(setting);
		break;
	default:
		break;
	}

}


/* ------ */

/* Writing */


static void writeBool(BerylSetting * setting)
{
	GROUPKEY;
	gboolean val = setting->value.value.as_bool;

	gconf_client_set_bool(client, PATH, val, NULL);
	g_free(keyname);

}

static void writeInt(BerylSetting * setting)
{
	GROUPKEY;
	int val = setting->value.value.as_int;

	gconf_client_set_int(client, PATH, val, NULL);
	g_free(keyname);
	
}

static void writeFloat(BerylSetting * setting)
{
	GROUPKEY;
	float val = setting->value.value.as_float;

	gconf_client_set_float(client, PATH, val, NULL);
	
	g_free(keyname);

}

static void writeString(BerylSetting * setting)
{
	GROUPKEY;
	gchar *val = setting->value.value.as_string;

	gconf_client_set_string(client, PATH, val, NULL);
	g_free(keyname);
}

static void writeColor(BerylSetting * setting)
{


	GROUPKEY;
	int red = setting->value.value.as_color.color.red;
	int blue = setting->value.value.as_color.color.blue;
	int green = setting->value.value.as_color.color.green;
	int alpha = setting->value.value.as_color.color.alpha;

	gconf_client_set_int(client, g_strconcat(PATH, "/red", NULL), red, NULL);
	gconf_client_set_int(client, g_strconcat(PATH, "/blue", NULL), blue,
						 NULL);
	gconf_client_set_int(client, g_strconcat(PATH, "/green", NULL), green,
						 NULL);
	gconf_client_set_int(client, g_strconcat(PATH, "/alpha", NULL), alpha,
						 NULL);
	g_free(keyname);

}

static void setKey(BerylSetting * setting)
{
	GROUPKEY;
	gchar *nkeyname;
	gchar *pre;
	gboolean bool_val;

	if (beryl_setting_value_get_key_enabled(&setting->value, &bool_val) &&
		bool_val)
		pre = "";
	else
		pre = "#";
	nkeyname = g_strconcat(keyname, "_keyboard", NULL);
	gint ire;
	gchar *smod = g_strdup(""), *bnum = g_strdup("Any");

	if (beryl_setting_value_get_buttonmods(&setting->value, &ire))
	{
		g_free(smod);
		smod = mods_to_string(ire);
	}
	if (beryl_setting_value_get_button(&setting->value, &ire))
	{
		g_free(bnum);
		bnum = g_strdup_printf("Button%d", ire);
	}
	pre = g_strconcat(pre, smod, bnum, NULL);
	g_free(smod);
	g_free(bnum);
	gconf_client_set_string(client,
							g_strconcat("/apps/beryl/", groupname,"/", nkeyname,
										NULL), pre, NULL);
	g_free(nkeyname);
	g_free(pre);
	g_free(keyname);
}

static void setButton(BerylSetting * setting)
{
	GROUPKEY;
	gchar *nkeyname;
	gchar *pre;
        gboolean bool_val;


	if (beryl_setting_value_get_button_enabled(&setting->value, &bool_val) &&
		bool_val)
		pre = "";
	else
		pre = "#";
	nkeyname = g_strconcat(keyname, "_mouse", NULL);
	gint ire;
	gchar *smod = g_strdup(""), *bnum = g_strdup("Any");

	if (beryl_setting_value_get_buttonmods(&setting->value, &ire))
	{
		g_free(smod);
		smod = mods_to_string(ire);
	}
	if (beryl_setting_value_get_button(&setting->value, &ire))
	{
		g_free(bnum);
		bnum = g_strdup_printf("Button%d", ire);
	}
	pre = g_strconcat(pre, smod, bnum, NULL);
	g_free(smod);
	g_free(bnum);
	gconf_client_set_string(client,
							g_strconcat("/apps/beryl/", groupname,"/", nkeyname,
										NULL), pre, NULL);
	g_free(pre);
	g_free(nkeyname);
	g_free(keyname);
}

static void setEdgemask(BerylSetting * setting)
{
	GROUPKEY;
	gchar *nkeyname;
        int ema;
        nkeyname=g_strconcat(keyname,"_edge",NULL);

        if (beryl_setting_value_get_edgemask(&setting->value,&ema))
        {
		gconf_client_set_int(client,g_strconcat("/apps/beryl/", groupname,"/", nkeyname,
										NULL),ema,NULL);
        }
	g_free(nkeyname);
	g_free(keyname);
}

static void setBell(BerylSetting * setting)
{
	GROUPKEY;
	gchar *nkeyname;
	gboolean bool_val;
        nkeyname=g_strconcat(keyname,"_bell",NULL);
	if (beryl_setting_value_get_bell(&setting->value,&bool_val))
	{
		gconf_client_set_bool(client,g_strconcat("/apps/beryl/", groupname,"/", nkeyname,
										NULL),bool_val,NULL);
	}
	g_free(nkeyname);
	g_free(keyname);

}

static void writeBinding(BerylSetting * setting)
{

	GROUPKEY;
	gboolean bool_val;

	if (beryl_setting_get_can_set_key(setting, &bool_val) && bool_val)
	{
		setKey(setting);
	}
	if (beryl_setting_get_can_set_button(setting, &bool_val) && bool_val)
	{
		setButton(setting);
	}
        if (beryl_setting_get_can_set_edgemask(setting,&bool_val) && bool_val)
	{
		setEdgemask(setting);
	}
     	if (beryl_setting_get_can_set_bell(setting,&bool_val) && bool_val)
	{
		setBell(setting);
	}
	g_free(keyname);
}

static void copy_string(BerylSettingValue * value, gchar *** ptr)
{
    *(*ptr)=g_strdup(value->value.as_string);
    *ptr=*ptr+1;
}

static void copy_bool(BerylSettingValue * value, gboolean ** ptr)
{
    *(*ptr++)=value->value.as_bool;
}

static void writeList(BerylSetting * setting)
{
	gchar * groupname=setting->parent->name?setting->parent->name:"_"; 
        gchar * keyname=g_strconcat(setting->is_screen?"s_":"a_",setting->name,NULL);
        gsize len=g_slist_length(setting->value.value.as_list);
	
        switch(setting->info.for_list.list_of_type)
	{
		case BERYL_SETTING_TYPE_BOOL:
		{
			
			gconf_client_set_list(client,PATH,GCONF_VALUE_BOOL,setting->value.value.as_list,NULL);
		}
		break;
		case BERYL_SETTING_TYPE_INT:
		{
			gconf_client_set_list(client,PATH,GCONF_VALUE_INT,setting->value.value.as_list,NULL);
		}
		break;
		case BERYL_SETTING_TYPE_FLOAT:
		{
			gconf_client_set_list(client,PATH,GCONF_VALUE_FLOAT,setting->value.value.as_list,NULL);
		}
		break;
		case BERYL_SETTING_TYPE_STRING:
		{
			gconf_client_set_list(client,PATH,GCONF_VALUE_STRING,setting->value.value.as_list,NULL);
		}
		break;
		default:
			break;
	}

}

static void writeSwitch(BerylSetting * setting)
{
	switch (setting->type)
	{
	case BERYL_SETTING_TYPE_BOOL:
		writeBool(setting);
		break;
	case BERYL_SETTING_TYPE_INT:
		writeInt(setting);
		break;
	case BERYL_SETTING_TYPE_FLOAT:
		writeFloat(setting);
		break;
	case BERYL_SETTING_TYPE_STRING:
		writeString(setting);
		break;
	case BERYL_SETTING_TYPE_COLOR:
		writeColor(setting);
		break;
	case BERYL_SETTING_TYPE_BINDING:
		writeBinding(setting);
		break;
	case BERYL_SETTING_TYPE_LIST:
		writeList(setting);
		break;
	default:
		break;
	}

}




/* ------ */


static void gconf_send_reload_signal(BerylSettingsContext * context)
{
	backendTriggered=FALSE;	
	beryl_settings_send_reload_signal();
}
/* Beryl Settings API */

gboolean backend_init(BerylSettingsContext * context)
{
	g_type_init();
	client=gconf_client_get_default();
	gconf_client_add_dir(client,"/apps/beryl",GCONF_CLIENT_PRELOAD_RECURSIVE,NULL);
	g_signal_connect(client,"value_changed",gconf_send_reload_signal,context);
	gconf_client_notify_add(client,"/apps/beryl",gconf_send_reload_signal,context,NULL,NULL);
}

gboolean backend_fini(BerylSettingsContext * context)
{

}

gboolean read_init(BerylSettingsContext * context)
{
	
	return TRUE;
}

void read_done(BerylSettingsContext * context)
{
}

gboolean write_init(BerylSettingsContext * context)
{
	

	return TRUE;
}

void write_done(BerylSettingsContext * context)
{
	backendTriggered=TRUE;
}

void read_setting(BerylSettingsContext * context, BerylSetting * setting)
{
	GROUPKEY;
	if (!gconf_client_get(client,PATH,NULL))
	{
		return;
	}
	
	setting->is_default = FALSE;
	readSwitch(setting);
}

void write_setting(BerylSettingsContext * context, BerylSetting * setting)
{
	if (backendTriggered)
	{
		GROUPKEY;
		if (setting->is_default)
		{
			return;
		}

		writeSwitch(setting);
	}
}

gboolean get_setting_is_integrated(BerylSetting * setting)
{
	return FALSE;
}

gboolean get_setting_is_read_only(BerylSetting * setting)
{
	return FALSE;
}

GSList *get_existing_profiles(void)
{
	GSList * ret = g_slist_alloc();
	return ret;
}
